//! d3-composer v0.6.4 - https://github.com/CSNW/d3-composer - @license: MIT
import { extent } from 'd3-array';
import { local } from 'd3-selection';
import { line, area, symbol, symbolCircle } from 'd3-shape';
import { axisTop, axisRight, axisBottom, axisLeft } from 'd3-axis';

function passthrough(d) {
  return d || [];
}

function array(value) {
  return typeof value === 'function'
    ? function(d, i, j) {
        return [value.call(this, d, i, j)];
      }
    : [value];
}

function fn(value) {
  return typeof value === 'function' ? value : () => value;
}

function seriesExtent(data, values, value) {
  return data.reduce(
    ([min, max], series, index) => {
      const [series_min, series_max] = extent(
        values(series, index, data),
        value
      );

      return [
        min != null && min < series_min ? min : series_min,
        max != null && max > series_max ? max : series_max
      ];
    },
    [undefined, undefined]
  );
}

const Size = local();
const Area = local();
const Series = local();

const zero_margin = [0, 0, 0, 0];

function toMargin(margin) {
  if (!margin) {
    return zero_margin;
  } else if (typeof margin === 'number') {
    return [margin, margin, margin, margin];
  } else if (!Array.isArray(margin)) {
    return [
      margin.top || 0,
      margin.right || 0,
      margin.bottom || 0,
      margin.left || 0
    ];
  } else {
    return margin;
  }
}

function toStyle(value) {
  if (!value) {
    return null;
  }
  if (typeof value === 'function') {
    return function() {
      return toStyle(value.apply(this, arguments));
    };
  }
  if (typeof value === 'string') {
    return value;
  }

  const { fixed, dynamic } = Object.keys(value).reduce(
    (memo, key) => {
      if (typeof value[key] !== 'function') {
        memo.fixed.push(`${key}: ${value[key]};`);
      } else {
        memo.dynamic.push(function(d, i, j) {
          return `${key}: ${value[key].call(this, d, i, j)};`;
        });
      }

      return memo;
    },
    { fixed: [], dynamic: [] }
  );

  if (!dynamic.length) {
    return fixed.join(' ');
  }

  return function(d, i, j) {
    return fixed.concat(dynamic.map(fn => fn.call(this, d, i, j))).join(' ');
  };
}

const PARSE_ROW = /\"(.*?)\"\s*([\w\d]+)/g;
const PARSE_MINMAX = /minmax\(([\w\d\s]+),([\w\d\s]+)\)/g;
const PARSE_REPEAT = /repeat\((\d+),([^\)]+)\)/g;

function parseTemplate(spec) {
  const formatted = spec
    .split('\n')
    .map(trim)
    .filter(Boolean)
    .join(' ');
  const [areas_rows_spec, columns_spec] = formatted.split('/', 2).map(trim);

  if (!areas_rows_spec || !columns_spec) {
    throw new Error('Invalid grid template');
  }

  let areas_spec = '';
  let rows = [];
  let match;
  while ((match = PARSE_ROW.exec(areas_rows_spec))) {
    const [_, area$$1, row] = match;

    areas_spec += `"${area$$1}"`;
    rows.push(toLength(row));
  }

  let areas;
  if (!areas_spec) {
    rows = parseTracks(areas_rows_spec);
    areas = {};
  } else {
    areas = parseAreas(areas_spec);
  }

  const columns = parseTracks(columns_spec);

  return { areas, rows, columns };
}

function parseTracks(spec) {
  let minmax_match;
  while ((minmax_match = PARSE_MINMAX.exec(spec))) {
    const [result, min, max] = minmax_match;

    // To handle potential nesting in repeat()
    // replace in spec with value that can be parsed by toLength
    const { index: start } = minmax_match;
    const end = start + result.length;
    const replacement = `MINMAX:${trim(min)}-${trim(max)}`;
    spec = spec.substr(0, start) + replacement + spec.substr(end);
  }

  let repeat_match;
  while ((repeat_match = PARSE_REPEAT.exec(spec))) {
    const [result, count, value] = repeat_match;
    const { index: start } = repeat_match;
    const end = start + result.length;

    const replacement = Array(parseInt(count, 10) + 1)
      .join(`${value} `)
      .trim();

    spec = spec.substr(0, start) + replacement + spec.substr(end);
  }

  const tracks = spec
    .split(' ')
    .filter(Boolean)
    .map(toLength);

  return tracks;
}

function parseAreas(spec) {
  const areas = {};

  spec
    .split('"')
    .filter((_, i) => i % 2 !== 0)
    .forEach((row_spec, row, rows) => {
      const columns = row_spec.split(' ').filter(Boolean);

      columns.forEach((name, column) => {
        if (name === '.') return;

        const next_row = rows[row + 1];
        const next_column = columns[column + 1];

        if (!areas[name]) {
          areas[name] = { top: row };
        }
        if (!('left' in areas[name])) {
          areas[name].left = column;
        }
        if (!next_row || next_row[column] !== name) {
          areas[name].bottom = row;
        }
        if (!next_column || next_column !== name) {
          areas[name].right = column;
        }
      });
    });

  return areas;
}

function toLength(value) {
  // units = 'fr' | 'percentage' | 'scalar'

  if (value.indexOf('MINMAX') === 0) {
    const [_, range] = value.split(':');
    const [min, max] = range.split('-');

    return [toLength(min), toLength(max)];
  }
  if (value === 'auto') {
    // For css grid, fr + auto would shrink auto to content
    // doesn't work well for this case
    // -> treat auto = 1fr
    return { value: 1, units: 'fr' };
  }
  if (value.endsWith('fr')) {
    return { value: Number(value.slice(0, -2)), units: 'fr' };
  }
  if (value.endsWith('%')) {
    return { value: Number(value.slice(0, -1)) / 100, units: 'percentage' };
  }

  try {
    return { value: Number(value), units: 'scalar' };
  } catch (err) {
    throw new Error(`Unknown length value "${value}"`);
  }
}

function trim(value) {
  return value.trim();
}

/*

See complete spec at https://drafts.csswg.org/css-grid/#algo-track-sizing

1. Initialize track sizes
  - base size, fixed = # or percentage
  - base size, flexible (fr or auto) = 0
  - growth limit, fixed = # or percentage
  - growth limit, flexible = base size
  - If growth limit is less than base size, increase growth limit to base size

2. (skipped, no intrinsic track sizes in @d3-composer/grid)

3. Maximize fixed tracks
  - S = grid size
  - B = combined base sizes
  - G = combined growth limits
  - A = available space
  - x = scaling factor
  
  S = Gx + A

  If G > S, A = 0 and x = S / G
  Otherwise, x = 1

4. Expand flexible tracks
  - A = available space
  - F = sum of flex factors

  fr = A / F

  If calculated fr is less than track base size, re-calculate fr, treating track as inflexible

5. (skipped, auto = 1fr in @d3-composer/grid)

*/

function solve(lengths, size, gap = 0, offset = 0) {
  size -= gap * (lengths.length - 1);

  // 1. Initialize track sizes
  let base_total = 0;
  let growth_total = 0;

  lengths = lengths.map(length => {
    let flexible = false;
    let base = 0;
    let growth = 0;

    if (Array.isArray(length)) {
      if (length[0].units === 'percentage') {
        base = length[0].value * size;
      } else if (length[0].units === 'scalar') {
        base = length[0].value;
      } else {
        throw new Error(`Invalid units "${length[0].units}" for base value`);
      }

      if (length[1].units === 'scalar') {
        growth = length[1].value;
      } else if (length[1].units === 'percentage') {
        growth = length[1].value * size;
      } else {
        growth = length[1].value;
        flexible = true;
      }
    } else {
      if (length.units === 'scalar') {
        base = growth = length.value;
      } else if (length.units === 'percentage') {
        base = growth = length.value * size;
      } else {
        base = 0;
        growth = length.value;
        flexible = true;
      }
    }

    base_total += base;
    growth_total += flexible ? 0 : growth;

    return { base, growth, flexible };
  });

  if (base_total > size) {
    throw new Error('More than 100% of available grid space was used');
  }

  // 2. (skipped)

  // 3. Maximize fixed tracks
  //
  // > If the free space is positive, distribute it equally to the base sizes of all tracks,
  // > freezing tracks as they reach their growth limits (and continuing to grow the unfrozen tracks as needed).
  //
  // Calculate a minimal step size so that the tracks increase equally

  if (growth_total > size) {
    let growth_size = base_total;
    let growing = lengths.filter(
      length => !length.flexible && length.growth > length.base
    );
    let available = size - growth_size;
    let step_size = available / growing.length;

    while (growth_size < size) {
      growing = growing.filter(length => {
        const remaining = length.growth - length.base;

        const step =
          available < remaining && available < step_size
            ? available
            : remaining < step_size
              ? remaining
              : step_size;

        length.base += step;
        growth_size += step;
        available -= step;

        return length.growth > length.base;
      });
      step_size = available / growing.length;
    }
  } else {
    // Growth is within free space, apply max growth to all
    lengths.forEach(length => {
      if (length.flexible) return;
      length.base = length.growth;
    });
  }

  // 4. Expand flexible tracks
  //
  // Calculate fr as available size / total factors
  // and apply unless growth is less than base,
  // in which case that track is made "inflexible"
  // and fr is re-calculated

  const available = growth_total > size ? 0 : size - growth_total;
  if (available) {
    let flexible = lengths.filter(({ flexible }) => flexible);
    let solved = flexible.length === 0;

    while (!solved) {
      const factor_total = flexible.reduce(
        (memo, { growth }) => memo + growth,
        0
      );
      const fr = available / factor_total;

      const valid = flexible.filter(length => {
        const { base, growth } = length;
        const value = growth * fr;

        if (value < base) {
          length.flexible = false;
          return false;
        } else {
          // If lengths are made inflexible,
          // fr is guaranteed to go up and base will increase
          // -> safe to set base here

          length.base = value;
          return true;
        }
      });

      if (valid.length === flexible.length) {
        solved = true;
      } else {
        flexible = valid;
      }
    }
  }

  // 5. (skipped)

  // Finally, extract [left, right] values for each track
  return lengths.reduce((memo, length, index) => {
    const last = memo[index - 1];

    const left = last ? last[1] + gap : offset;
    const right = left + length.base;

    memo.push([left, right]);
    return memo;
  }, []);
}

const cache = new Map();

function template(spec, options) {
  let { width, height, gap, row_gap, column_gap, padding } = options;
  padding = toMargin(padding);
  width -= padding[1] + padding[3];
  height -= padding[0] + padding[2];
  if (gap) row_gap = column_gap = gap;

  // spec: string = template
  // spec: object = { rows, columns, [areas] } templates
  let rows, columns, areas;
  if (typeof spec === 'object') {
    rows = parseTracks(spec.rows);
    columns = parseTracks(spec.columns);
    areas = spec.areas ? parseAreas(spec.areas) : {};
  } else {
    const parsed = cache.has(spec) ? cache.get(spec) : parseTemplate(spec);
    cache.set(spec, parsed);
    ({ rows, columns, areas } = parsed);
  }

  const x = solve(columns, width, column_gap, padding[3]);
  const y = solve(rows, height, row_gap, padding[0]);

  return layout(rows, columns, areas, x, y);
}

function layout(rows, columns, areas, x, y) {
  const result = { rows: y, columns: x, items: [], areas: {} };

  for (let i = 0; i < rows.length; i++) {
    for (let j = 0; j < columns.length; j++) {
      const top = y[i][0];
      const right = x[j][1];
      const bottom = y[i][1];
      const left = x[j][0];

      result.items.push({
        x: left,
        y: top,
        width: right - left,
        height: bottom - top,
        top,
        right,
        bottom,
        left
      });
    }
  }

  for (const name of Object.keys(areas)) {
    const area$$1 = areas[name];

    const top = y[area$$1.top][0];
    const right = x[area$$1.right][1];
    const bottom = y[area$$1.bottom][1];
    const left = x[area$$1.left][0];

    result.areas[name] = {
      x: left,
      y: top,
      width: right - left,
      height: bottom - top,
      top,
      right,
      bottom,
      left
    };
  }

  return result;
}

function applyMargin(area$$1, margin) {
  margin = toMargin(margin);

  let { top, right, bottom, left } = area$$1;
  top += margin[0];
  right -= margin[1];
  bottom -= margin[2];
  left += margin[3];

  return {
    x: left,
    y: top,
    width: right - left,
    height: bottom - top,
    top,
    right,
    bottom,
    left
  };
}

const size_id = Size.toString();

/**
 * Determine node size for given selection
 *
 * 1. Check viewBox (if svg)
 * 2. Check attributes (if svg)
 * 3. Check Size local
 * 4. Check bbox
 *
 * @param {d3.selection} selection
 */
function size(selection) {
  const node = selection.node();

  if (node.tagName === 'svg') {
    const viewBox = node.getAttribute('viewBox');
    if (viewBox) {
      const [_1, _2, width, height] = viewBox.split(' ');
      return { width: Number(width), height: Number(height) };
    }

    const width = node.getAttribute('width');
    const height = node.getAttribute('height');

    if (width != null && height != null) {
      return { width: Number(width), height: Number(height) };
    }
  }

  if (size_id in node) {
    return node[size_id];
  }

  try {
    const bbox = node.getBBox();
    return { width: bbox.width, height: bbox.height };
  } catch (err) {
    return { width: NaN, height: NaN };
  }
}

function childNodes() {
  return this.childNodes;
}

function interpolatePath(selection, path, interpolate) {
  if (interpolate) {
    selection.attrTween('d', function(d, i, j) {
      const next = typeof path === 'function' ? path.call(this, d, i, j) : path;
      const previous = this.getAttribute('d');

      return interpolate(previous, next);
    });
  } else {
    selection.attr('d', path);
  }
}

function translateXY(_x, _y) {
  return function(d, i, j) {
    const x = _x.call(this, d, i, j);
    const y = _y.call(this, d, i, j);

    return `translate(${x}, ${y})`;
  };
}

function chart(selection, options = {}) {
  const { width = 400, height = 300, responsive = false } = options;

  let svg;
  if (selection.node().tagName === 'svg') {
    svg = selection;
  } else {
    const ratio = (height / width) * 100;
    const style = responsive
      ? `position: relative; width: 100%; height: 0; padding-top: ${ratio}%;`
      : null;

    selection.attr('data-container', '').attr('style', style);
    svg = svgLayer(selection, { responsive });
  }

  return svg
    .attr('viewBox', responsive ? `0 0 ${width} ${height}` : null)
    .attr('preserveAspectRatio', responsive ? 'xMidYMid meet' : null)
    .attr('width', responsive ? null : width)
    .attr('height', responsive ? null : height);
}

function svgLayer(selection, options) {
  const { responsive } = options;

  const style = responsive ? 'position: absolute; top: 0; left: 0;' : null;

  const svg = selection
    .selectAll(childNodes)
    .filter('svg')
    .data([null]);

  return svg
    .enter()
    .append('svg')
    .merge(svg)
    .attr('style', style);
}

function layer(selection, id, options = {}) {
  const { element = 'g' } = options;
  const instance = selection
    .selectAll(childNodes)
    .filter(`[data-layer="${id}"]`)
    .data([null]);

  // Forward size from selection to layer
  const size$$1 = selection.property(Size);

  return instance
    .enter()
    .append(element)
    .attr('data-layer', id)
    .merge(instance)
    .property(Size, size$$1);
}

function layer$1(selection, id, area$$1, options) {
  const { margin, style, class: className } = options;
  area$$1 = applyMargin(area$$1, margin);

  return layer(selection, id)
    .property(Area, area$$1)
    .property(Size, area$$1)
    .attr('style', toStyle(style))
    .attr('class', className)
    .attr('transform', `translate(${area$$1.x}, ${area$$1.y})`);
}

function layout$1(selection, grid, callback) {
  let index = 0;
  const indices = {};
  const added = [];

  function layers(id, options = {}) {
    if (id && typeof id === 'object') {
      options = id;
      id = undefined;
    }

    const item = index++;
    if (index >= grid.items.length) index = 0;

    const name = `__item_${item}`;
    if (!indices[name]) indices[name] = 0;
    const i = indices[name]++;

    id = id || `item-${item}${i > 0 ? `-${i}` : ''}`;
    added.push(id);

    return layer$1(selection, id, grid.items[item], options).attr(
      'data-item',
      item
    );
  }

  Object.keys(grid.areas).forEach(name => {
    const area$$1 = grid.areas[name];
    indices[name] = 0;

    layers[name] = (id, options = {}) => {
      if (id && typeof id === 'object') {
        options = id;
        id = undefined;
      }
      const i = indices[name]++;

      id = id || `area-${name}${i > 0 ? `-${i}` : ''}`;
      added.push(id);

      return layer$1(selection, id, area$$1, options).attr('data-area', name);
    };
  });

  // TODO This isn't ideomatic d3, but not sure the best way to deal with dynamic layers
  callback(layers);

  // Remove stale layers
  selection
    .selectAll(childNodes)
    .filter('[data-layer]')
    .data(added, function(d) {
      return d || this.dataset.layer;
    })
    .order()
    .exit()
    .remove();

  return selection;
}

const seriesKey = (d, i) => (d && d.key != null ? d.key : i);

function series(selection, options) {
  const {
    data = passthrough,
    key = seriesKey,
    style,
    class: className
  } = options;

  const layers = selection
    .selectAll(childNodes)
    .filter('[data-series]')
    .data(data, key);

  layers.exit().remove();

  return layers
    .enter()
    .append('g')
    .attr('data-series', key)
    .property(Series, d => d)
    .merge(layers)
    .attr('class', className)
    .attr('style', toStyle(style))
    .order();
}

function line$1(selection, options = {}) {
  const {
    data = passthrough,
    x,
    y,
    key,
    style,
    class: className,
    transition,
    curve,
    interpolate,
    defined
  } = options;

  const line$$1 = line()
    .x(x)
    .y(y);
  if (curve) line$$1.curve(curve);
  if (defined) line$$1.defined(defined);

  const path = selection.selectAll('path').data(array(data), key);

  path.exit().remove();

  path
    .enter()
    .append('path')
    .attr('d', line$$1)
    .merge(path)
    .attr('fill', 'none')
    .attr('stroke', 'currentColor')
    .attr('style', toStyle(style))
    .attr('class', className)
    .transition(transition)
    .call(interpolatePath, line$$1, interpolate);

  return selection;
}

function bars(selection, options) {
  let {
    data = passthrough,
    x,
    x0,
    x1,
    y,
    y0,
    y1,
    width,
    height,
    key,
    style,
    class: className,
    transition
  } = options;

  x = x != null ? x : x0;
  y = y != null ? y : y1;

  width =
    width ||
    function(d, i, j) {
      return Math.abs(fn(x1).call(this, d, i, j) - fn(x).call(this, d, i, j));
    };
  height =
    height ||
    function(d, i, j) {
      return Math.abs(fn(y).call(this, d, i, j) - fn(y0).call(this, d, i, j));
    };

  const bars = selection.selectAll('rect').data(data, key);

  bars.exit().remove();
  bars
    .enter()
    .append('rect')
    .attr('x', x)
    .attr('y', y)
    .attr('height', height)
    .attr('width', width)
    .merge(bars)
    .attr('fill', 'currentColor')
    .attr('stroke', 'none')
    .attr('style', toStyle(style))
    .attr('class', className)
    .transition(transition)
    .attr('x', x)
    .attr('y', y)
    .attr('height', height)
    .attr('width', width);

  return selection;
}

function scatter(selection, options) {
  const {
    data = passthrough,
    x,
    y,
    key,
    path,
    style,
    class: className,
    transition,
    interpolate
  } = options;
  const translate = translateXY(x, y);

  const paths = selection.selectAll('path').data(data, key);

  paths.exit().remove();

  paths
    .enter()
    .append('path')
    .attr('transform', translate)
    .merge(paths)
    .attr('fill', 'currentColor')
    .attr('stroke', 'none')
    .attr('style', toStyle(style))
    .attr('class', className)
    .transition(transition)
    .attr('transform', translate)
    .call(interpolatePath, path, interpolate);

  return selection;
}

function area$1(selection, options) {
  const {
    data = passthrough,
    x,
    x0,
    x1,
    y0,
    y1,
    key,
    style,
    class: className,
    transition,
    curve,
    interpolate,
    defined
  } = options;

  const area$$1 = area()
    .y0(y0)
    .y1(y1);
  if (x) area$$1.x(x);
  if (x0 && x1) area$$1.x0(x0).x1(x1);
  if (curve) area$$1.curve(curve);
  if (defined) area$$1.defined(defined);

  const path = selection.selectAll('path').data(array(data), key);

  path
    .exit()
    .transition(transition)
    .remove();
  path
    .enter()
    .append('path')
    .attr('d', area$$1)
    .merge(path)
    .attr('fill', 'currentColor')
    .attr('stroke', 'none')
    .attr('style', toStyle(style))
    .attr('class', className)
    .transition(transition)
    .call(interpolatePath, area$$1, interpolate);

  return selection;
}

function labels(selection, options) {
  const {
    data = passthrough,
    x,
    y,
    key,
    text,
    style,
    class: className,
    transition,
    transform,
    anchor = 'start',
    baseline = 'hanging'
  } = options;
  const translate = translateXY(x, y);

  const groups = selection.selectAll('g').data(data, key);

  groups.exit().remove();

  const entering = groups
    .enter()
    .append('g')
    .attr('transform', translate);

  entering
    .merge(groups)
    .transition(transition)
    .attr('transform', translate);
  entering
    .append('text')
    .merge(groups.select('text'))
    .attr('text-anchor', anchor)
    .attr('dominant-baseline', baseline)
    .attr('style', toStyle(style))
    .attr('class', className)
    .attr('transform', transform || null)
    .text(text);

  return selection;
}

function axisTop$1(selection, options) {
  const {
    scale,
    xScale,
    style,
    domainStyle,
    tickStyle,
    labelStyle,
    class: className,
    transition
  } = options;
  const axis = prepare(axisTop(scale || xScale), options);
  const { height } = size(selection);

  layer(selection, 'axis-transform')
    .attr('transform', `translate(0, ${height || 0})`)
    .attr('class', className)
    .transition(transition)
    .call(axis)
    .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

  return selection;
}

function axisRight$1(selection, options) {
  const {
    scale,
    yScale,
    style,
    domainStyle,
    tickStyle,
    labelStyle,
    class: className,
    transition
  } = options;
  const axis = prepare(axisRight(scale || yScale), options);

  selection
    .attr('class', className)
    .transition(transition)
    .call(axis)
    .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

  return selection;
}

function axisBottom$1(selection, options) {
  const {
    scale,
    xScale,
    style,
    domainStyle,
    tickStyle,
    labelStyle,
    class: className,
    transition
  } = options;
  const axis = prepare(axisBottom(scale || xScale), options);

  selection
    .attr('class', className)
    .transition(transition)
    .call(axis)
    .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

  return selection;
}

function axisLeft$1(selection, options) {
  const {
    scale,
    yScale,
    style,
    domainStyle,
    tickStyle,
    labelStyle,
    class: className,
    transition
  } = options;
  const axis = prepare(axisLeft(scale || yScale), options);
  const { width } = size(selection);

  layer(selection, 'axis-transform')
    .attr('transform', `translate(${width || 0}, 0)`)
    .attr('class', className)
    .transition(transition)
    .call(axis)
    .call(styleAxis, style, domainStyle, tickStyle, labelStyle);

  return selection;
}

function prepare(axis, options) {
  const {
    ticks,
    tickArguments,
    tickValues,
    tickFormat,
    tickSize,
    tickSizeInner,
    tickSizeOuter,
    tickPadding
  } = options;

  if (ticks) axis.ticks(ticks);
  if (tickArguments) axis.tickArguments(tickArguments);
  if (tickValues) axis.tickValues(tickValues);
  if (tickFormat) axis.tickFormat(tickFormat);
  if (tickSize) axis.tickSize(tickSize);
  if (tickSizeInner) axis.tickSizeInner(tickSizeInner);
  if (tickSizeOuter) axis.tickSizeOuter(tickSizeOuter);
  if (tickPadding) axis.tickPadding(tickPadding);

  return function(selection) {
    axis.call(this, selection);

    // Reset hard-coded text values
    selection.attr('font-size', null).attr('font-family', null);
  };
}

function styleAxis(selection, style, domainStyle, tickStyle, labelStyle) {
  // Get underlying selection from transition
  selection = selection.selection ? selection.selection() : selection;

  selection.attr('style', toStyle(style));
  selection.select('.domain').attr('style', toStyle(domainStyle));
  selection.selectAll('.tick line').attr('style', toStyle(tickStyle));
  selection.selectAll('.tick text').attr('style', toStyle(labelStyle));
}

function text(selection, options) {
  // anchor = origin-x = 'start' | 'middle' | 'end'
  //   (https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/text-anchor)
  // baseline = origin-y = 'hanging' | 'middle' | 'baseline'
  //   (https://developer.mozilla.org/en-US/docs/Web/SVG/Attribute/dominant-baseline)
  // justify = container-x = 'start' | 'center' | 'end'
  //   (https://css-tricks.com/snippets/css/complete-guide-grid/#article-header-id-30)
  // align = container-y = 'start' | 'center' | 'end'
  //   (https://css-tricks.com/snippets/css/complete-guide-grid/#article-header-id-31)

  let {
    text,
    anchor,
    baseline,
    justify = 'start',
    align = 'start',
    class: className,
    style,
    rotation,
    transform
  } = options;

  anchor =
    anchor != null
      ? anchor
      : justify === 'center'
        ? 'middle'
        : justify === 'end'
          ? 'end'
          : 'start';

  baseline =
    baseline != null
      ? baseline
      : align === 'center'
        ? 'middle'
        : align === 'end'
          ? 'baseline'
          : 'hanging';

  if (!transform && (justify !== 'start' || align !== 'start' || rotation)) {
    const { width, height } = size(selection);
    let x = 0;
    let y = 0;

    if (justify === 'center') {
      x = width / 2;
    } else if (justify === 'end') {
      x = width;
    }

    if (align === 'center') {
      y = height / 2;
    } else if (align === 'end') {
      y = height;
    }

    transform = `translate(${x}, ${y}) rotate(${rotation || 0})`;
  }

  const instance = selection.selectAll('text').data([null]);
  instance
    .enter()
    .append('text')
    .merge(instance)
    .attr('dominant-baseline', baseline)
    .attr('text-anchor', anchor)
    .attr('transform', transform)
    .attr('class', className)
    .attr('style', toStyle(style))
    .text(text);

  return selection;
}

function legend(selection, options) {
  let {
    path,
    size,
    width,
    height,
    text = d => d,
    data,
    groupStyle,
    groupClass,
    pathStyle,
    labelStyle
  } = options;

  // TODO Accept size, width, or height as functions
  if (!height && !size) {
    size = 50;
  }
  if (!height) {
    height = 3 * Math.sqrt(size);
  }
  if (!width) {
    width = height;
  }
  if (!size) {
    size = (Math.min(width, height) * Math.min(width, height)) / 3;
  }

  if (!path) {
    path = symbol()
      .size(size)
      .type(symbolCircle);
  }

  const layers = selection
    .selectAll('g')
    .data(data, (d, i) => (d && d.key != null ? d.key : i));

  layers.exit().remove();

  const entering = layers.enter().append('g');

  entering
    .merge(layers)
    .attr('transform', (_, i) => `translate(0, ${i * height})`)
    .attr('style', toStyle(groupStyle))
    .attr('class', groupClass);
  entering
    .append('path')
    .merge(layers.select('path'))
    .attr('transform', `translate(${width / 2}, ${height / 2})`)
    .attr('fill', 'currentColor')
    .attr('stroke', 'none')
    .attr('style', toStyle(pathStyle))
    .attr('d', path);
  entering
    .append('text')
    .merge(layers.select('text'))
    .attr('dominant-baseline', 'central')
    .attr('transform', `translate(${width}, ${height / 2})`)
    .attr('fill', 'currentColor')
    .attr('stroke', 'none')
    .attr('style', toStyle(labelStyle))
    .text(text);

  return selection;
}

const index = (_, i) => i;

function gridlines(selection, options) {
  const {
    xGrid = true,
    yGrid = true,
    xScale,
    yScale,
    xTicks,
    xTickArguments,
    xTickValues,
    yTicks,
    yTickArguments,
    yTickValues,
    style,
    class: className,
    transition
  } = options;

  const [x1, x2] = xScale.range();
  const [y1, y2] = yScale.range();

  const translateX = d => `translate(${xScale(d) + 0.5}, 0)`;
  const translateY = d => `translate(0, ${yScale(d) + 0.5})`;

  const x_data = !xGrid
    ? []
    : xTickValues ||
      xScale.ticks.apply(
        xScale.ticks,
        xTicks ? [xTicks] : xTickArguments || []
      );
  const y_data = !yGrid
    ? []
    : yTickValues ||
      yScale.ticks.apply(
        yScale.ticks,
        yTicks ? [yTicks] : yTickArguments || []
      );

  const x_lines = layer(selection, 'x')
    .selectAll('line')
    .data(x_data, index);

  x_lines.exit().remove();
  x_lines
    .enter()
    .append('line')
    .attr('transform', translateX)
    .merge(x_lines)
    .attr('x1', 0)
    .attr('x2', 0)
    .attr('y1', y1)
    .attr('y2', y2)
    .attr('style', toStyle(style))
    .attr('class', className)
    .transition(transition)
    .attr('transform', translateX);

  const y_lines = layer(selection, 'y')
    .selectAll('line')
    .data(y_data, index);

  y_lines.exit().remove();
  y_lines
    .enter()
    .append('line')
    .attr('transform', translateY)
    .merge(y_lines)
    .attr('x1', x1)
    .attr('x2', x2)
    .attr('y1', 0)
    .attr('y2', 0)
    .attr('style', toStyle(style))
    .attr('class', className)
    .transition(transition)
    .attr('transform', translateY);

  return selection;
}

const symbolLine = {
  draw(context, size) {
    const w = Math.sqrt(size) * 2;
    const h = size / w;
    const x = -w / 2;
    const y = -h / 2;

    context.rect(x, y, w, h);
  }
};

export { passthrough, array, seriesExtent, toStyle, toMargin, Size, Area, Series, template, applyMargin, chart, layout$1 as layout, layer, series, line$1 as line, bars, scatter, area$1 as area, labels, axisLeft$1 as axisLeft, axisRight$1 as axisRight, axisBottom$1 as axisBottom, axisTop$1 as axisTop, text, legend, gridlines, symbolLine, size, interpolatePath, translateXY };
